/*
 *  linux/drivers/mmc/core/mmc_ops.h
 *
 *  Copyright 2006-2007 Pierre Ossman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 */

#include <linux/types.h>
#include <linux/scatterlist.h>

#include <linux/mmc1/host.h>
#include <linux/mmc1/card.h>
#include <linux/mmc1/mmc.h>

#include "core.h"
#include "mmc_ops.h"

#if 0
#define DBG(x...)	printk(KERN_ALERT x)
#else
#define DBG(x...)	do { } while (0)
#endif

static int _mmc1_select_card(struct mmc_host *host, struct mmc_card *card)
{
	int err;
	struct mmc_command cmd;
	DBG("[%s] s\n",__func__);

	BUG_ON(!host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SELECT_CARD;

	if (card) {
		cmd.arg = card->rca << 16;
//		cmd.flags = MMC_RSP_R1 | MMC_CMD_AC;	//zhf: modified by James Tian
		cmd.flags = MMC_RSP_R1B | MMC_CMD_AC;
	} else {
		cmd.arg = 0;
		cmd.flags = MMC_RSP_NONE | MMC_CMD_AC;
	}

	err = mmc1_wait_for_cmd(host, &cmd, MMC_CMD_RETRIES);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}
	DBG("[%s] e2\n",__func__);
	return 0;
}

int mmc1_select_card(struct mmc_card *card)
{
	int ret;
	
	DBG("[%s] s\n",__func__);
	BUG_ON(!card);
	ret = _mmc1_select_card(card->host, card);

	DBG("[%s] e\n",__func__);
	return ret;
}

int mmc1_deselect_cards(struct mmc_host *host)
{
	int ret;
	DBG("[%s] s\n",__func__);
	ret = _mmc1_select_card(host, NULL);
	DBG("[%s] e\n",__func__);
	return ret;
}

int mmc1_card_sleepawake(struct mmc_host *host, int sleep)
{
	struct mmc_command cmd;
	struct mmc_card *card = host->card;
	int err;
	
	DBG("[%s] s\n",__func__);
	
	if (sleep)
		mmc1_deselect_cards(host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SLEEP_AWAKE;
	cmd.arg = card->rca << 16;
	if (sleep)
		cmd.arg |= 1 << 15;

	cmd.flags = MMC_RSP_R1B | MMC_CMD_AC;
	err = mmc1_wait_for_cmd(host, &cmd, 0);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}

	/*
	 * If the host does not wait while the card signals busy, then we will
	 * will have to wait the sleep/awake timeout.  Note, we cannot use the
	 * SEND_STATUS command to poll the status because that command (and most
	 * others) is invalid while the card sleeps.
	 */
	if (!(host->caps & MMC_CAP_WAIT_WHILE_BUSY))
		mmc1_delay(DIV_ROUND_UP(card->ext_csd.sa_timeout, 10000));

	if (!sleep)
		err = mmc1_select_card(card);

	DBG("[%s] e2\n",__func__);
	return err;
}

int mmc1_go_idle(struct mmc_host *host)
{
	int err;
	struct mmc_command cmd;

	DBG("[%s] s\n",__func__);
	/*
	 * Non-SPI hosts need to prevent chipselect going active during
	 * GO_IDLE; that would put chips into SPI mode.  Remind them of
	 * that in case of hardware that won't pull up DAT3/nCS otherwise.
	 *
	 * SPI hosts ignore ios.chip_select; it's managed according to
	 * rules that must accomodate non-MMC slaves which this layer
	 * won't even know about.
	 */
	if (!mmc1_host_is_spi(host)) {
		mmc1_set_chip_select(host, MMC_CS_HIGH);
		mmc1_delay(1);
	}

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_GO_IDLE_STATE;
	cmd.arg = 0;
//	cmd.flags = MMC_RSP_SPI_R1 | MMC_RSP_NONE | MMC_CMD_BC;
	cmd.flags = MMC_RSP_NONE | MMC_CMD_BC;	//zhf: mark SPI mode temporarily by James Tian

	err = mmc1_wait_for_cmd(host, &cmd, 0);

	mmc1_delay(1);

	if (!mmc1_host_is_spi(host)) {
		mmc1_set_chip_select(host, MMC_CS_DONTCARE);
		mmc1_delay(1);
	}

	host->use_spi_crc = 0;

	DBG("[%s] e\n",__func__);
	return err;
}

int mmc1_send_op_cond(struct mmc_host *host, u32 ocr, u32 *rocr)
{
	struct mmc_command cmd;
	int i, err = 0;
	DBG("[%s] s\n",__func__);

	BUG_ON(!host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SEND_OP_COND;
	cmd.arg = mmc1_host_is_spi(host) ? 0 : ocr;
//	cmd.flags = MMC_RSP_SPI_R1 | MMC_RSP_R3 | MMC_CMD_BCR;
	cmd.flags = MMC_RSP_R3 | MMC_CMD_BCR;	//zhf: mark SPI mode temporarily by James Tian

	for (i = 100; i; i--) {
		err = mmc1_wait_for_cmd(host, &cmd, 0);
		if (err)
			break;

		/* if we're just probing, do a single pass */
		if (ocr == 0)
			break;

		/* otherwise wait until reset completes */
		if (mmc1_host_is_spi(host)) {
			if (!(cmd.resp[0] & R1_SPI_IDLE))
				break;
		} else {
			if (cmd.resp[0] & MMC_CARD_BUSY)
				break;
		}

		err = -ETIMEDOUT;

		mmc1_delay(10);
	}

	if (rocr && !mmc1_host_is_spi(host))
		*rocr = cmd.resp[0];

	DBG("[%s] e\n",__func__);
	return err;
}

int mmc1_all_send_cid(struct mmc_host *host, u32 *cid)
{
	int err;
	struct mmc_command cmd;
	DBG("[%s] s\n",__func__);

	BUG_ON(!host);
	BUG_ON(!cid);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_ALL_SEND_CID;
	cmd.arg = 0;
	cmd.flags = MMC_RSP_R2 | MMC_CMD_BCR;

	err = mmc1_wait_for_cmd(host, &cmd, MMC_CMD_RETRIES);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}

	memcpy(cid, cmd.resp, sizeof(u32) * 4);

	DBG("[%s] e2\n",__func__);
	return 0;
}

int mmc1_set_relative_addr(struct mmc_card *card)
{
	int err;
	struct mmc_command cmd;
	DBG("[%s] s\n",__func__);

	BUG_ON(!card);
	BUG_ON(!card->host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SET_RELATIVE_ADDR;
	cmd.arg = card->rca << 16;
	cmd.flags = MMC_RSP_R1 | MMC_CMD_AC;

	err = mmc1_wait_for_cmd(card->host, &cmd, MMC_CMD_RETRIES);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}

	DBG("[%s] e2\n",__func__);
	return 0;
}

static int
mmc1_send_cxd_native(struct mmc_host *host, u32 arg, u32 *cxd, int opcode)
{
	int err;
	struct mmc_command cmd;
	DBG("[%s] s\n",__func__);
	
	BUG_ON(!host);
	BUG_ON(!cxd);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = opcode;
	cmd.arg = arg;
	cmd.flags = MMC_RSP_R2 | MMC_CMD_AC;

	err = mmc1_wait_for_cmd(host, &cmd, MMC_CMD_RETRIES);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}
	memcpy(cxd, cmd.resp, sizeof(u32) * 4);

	DBG("[%s] e2\n",__func__);
	return 0;
}

static int
mmc1_send_cxd_data(struct mmc_card *card, struct mmc_host *host,
		u32 opcode, void *buf, unsigned len)
{
	struct mmc_request mrq;
	struct mmc_command cmd;
	struct mmc_data data;
	struct scatterlist sg;
	void *data_buf;

	DBG("[%s] s\n",__func__);
	/* dma onto stack is unsafe/nonportable, but callers to this
	 * routine normally provide temporary on-stack buffers ...
	 */
	data_buf = kmalloc(len, GFP_KERNEL);
	if (data_buf == NULL) {
		DBG("[%s] e1\n",__func__);
		return -ENOMEM;
	}

	memset(&mrq, 0, sizeof(struct mmc_request));
	memset(&cmd, 0, sizeof(struct mmc_command));
	memset(&data, 0, sizeof(struct mmc_data));

	mrq.cmd = &cmd;
	mrq.data = &data;

	cmd.opcode = opcode;
	cmd.arg = 0;

	/* NOTE HACK:  the MMC_RSP_SPI_R1 is always correct here, but we
	 * rely on callers to never use this with "native" calls for reading
	 * CSD or CID.  Native versions of those commands use the R2 type,
	 * not R1 plus a data block.
	 */
//	cmd.flags = MMC_RSP_SPI_R1 | MMC_RSP_R1 | MMC_CMD_ADTC;
	cmd.flags = MMC_RSP_R1 | MMC_CMD_ADTC;	//zhf: mark SPI mode temporarily by James Tian

	data.blksz = len;
	data.blocks = 1;
	data.flags = MMC_DATA_READ;
	data.sg = &sg;
	data.sg_len = 1;

	sg_init_one(&sg, data_buf, len);

	if (opcode == MMC_SEND_CSD || opcode == MMC_SEND_CID) {
		/*
		 * The spec states that CSR and CID accesses have a timeout
		 * of 64 clock cycles.
		 */
		data.timeout_ns = 0;
		data.timeout_clks = 64;
	} else
		mmc1_set_data_timeout(&data, card);

	mmc1_wait_for_req(host, &mrq);

	memcpy(buf, data_buf, len);
	kfree(data_buf);

	if (cmd.error) {
		DBG("[%s] e2\n",__func__);
		return cmd.error;
	}
	if (data.error) {
		DBG("[%s] e3\n",__func__);
		return data.error;
	}

	DBG("[%s] e4\n",__func__);
	return 0;
}

int mmc1_send_csd(struct mmc_card *card, u32 *csd)
{
	int ret, i;
	DBG("[%s] s\n",__func__);

	if (!mmc1_host_is_spi(card->host)) {
		ret = mmc1_send_cxd_native(card->host, card->rca << 16,
				csd, MMC_SEND_CSD);
		DBG("[%s] e1\n",__func__);
		return ret;
	}

	ret = mmc1_send_cxd_data(card, card->host, MMC_SEND_CSD, csd, 16);
	if (ret) {
		DBG("[%s] e2\n",__func__);
		return ret;
	}

	for (i = 0;i < 4;i++)
		csd[i] = be32_to_cpu(csd[i]);

	DBG("[%s] e3\n",__func__);
	return 0;
}

int mmc1_send_cid(struct mmc_host *host, u32 *cid)
{
	int ret, i;
	DBG("[%s] s\n",__func__);
	if (!mmc1_host_is_spi(host)) {
		if (!host->card) {
			DBG("[%s] e1\n",__func__);
			return -EINVAL;
		}
		ret = mmc1_send_cxd_native(host, host->card->rca << 16,
				cid, MMC_SEND_CID);
		DBG("[%s] e2\n",__func__);
		return ret;
	}

	ret = mmc1_send_cxd_data(NULL, host, MMC_SEND_CID, cid, 16);
	if (ret) {
		DBG("[%s] e3\n",__func__);
		return ret;
	}

	for (i = 0;i < 4;i++)
		cid[i] = be32_to_cpu(cid[i]);

	DBG("[%s] e3\n",__func__);
	return 0;
}

int mmc1_send_ext_csd(struct mmc_card *card, u8 *ext_csd)
{
	int ret;

	DBG("[%s] s\n",__func__);
	ret = mmc1_send_cxd_data(card, card->host, MMC_SEND_EXT_CSD,
			ext_csd, 512);
	DBG("[%s] e\n",__func__);
	return ret;
}

int mmc1_spi_read_ocr(struct mmc_host *host, int highcap, u32 *ocrp)
{
	struct mmc_command cmd;
	int err;

	DBG("[%s] s\n",__func__);
	
	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SPI_READ_OCR;
	cmd.arg = highcap ? (1 << 30) : 0;
	cmd.flags = MMC_RSP_SPI_R3;

	err = mmc1_wait_for_cmd(host, &cmd, 0);

	*ocrp = cmd.resp[1];

	DBG("[%s] e\n",__func__);
	return err;
}

int mmc1_spi_set_crc(struct mmc_host *host, int use_crc)
{
	struct mmc_command cmd;
	int err;

	DBG("[%s] s\n",__func__);
	
	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SPI_CRC_ON_OFF;
	cmd.flags = MMC_RSP_SPI_R1;
	cmd.arg = use_crc;

	err = mmc1_wait_for_cmd(host, &cmd, 0);
	if (!err)
		host->use_spi_crc = use_crc;

	DBG("[%s] e\n",__func__);
	return err;
}

int mmc1_switch(struct mmc_card *card, u8 set, u8 index, u8 value)
{
	int err;
	struct mmc_command cmd;
	u32 status;
	DBG("[%s] s\n",__func__);

	BUG_ON(!card);
	BUG_ON(!card->host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SWITCH;
	cmd.arg = (MMC_SWITCH_MODE_WRITE_BYTE << 24) |
		  (index << 16) |
		  (value << 8) |
		  set;
//	cmd.flags = MMC_RSP_SPI_R1B | MMC_RSP_R1B | MMC_CMD_AC;
	cmd.flags = MMC_RSP_R1B | MMC_CMD_AC;	//zhf: mark SPI mode temporarily by James Tian

	err = mmc1_wait_for_cmd(card->host, &cmd, MMC_CMD_RETRIES);
	if (err)
		return err;

	/* Must check status to be sure of no errors */
	do {
		err = mmc1_send_status(card, &status);
		if (err)
			return err;
		if (card->host->caps & MMC_CAP_WAIT_WHILE_BUSY)
			break;
		if (mmc1_host_is_spi(card->host))
			break;
	} while (R1_CURRENT_STATE(status) == 7);

	if (mmc1_host_is_spi(card->host)) {
		if (status & R1_SPI_ILLEGAL_COMMAND)
			return -EBADMSG;
	} else {
		if (status & 0xFDFFA000)
			printk(KERN_WARNING "%s: unexpected status %#x after "
			       "switch", mmc1_hostname(card->host), status);
		if (status & R1_SWITCH_ERROR)
			return -EBADMSG;
	}

	DBG("[%s] e2\n",__func__);
	return 0;
}

int mmc1_send_status(struct mmc_card *card, u32 *status)
{
	int err;
	struct mmc_command cmd;

	DBG("[%s] s\n",__func__);
	
	BUG_ON(!card);
	BUG_ON(!card->host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = MMC_SEND_STATUS;
	if (!mmc1_host_is_spi(card->host))
		cmd.arg = card->rca << 16;
//	cmd.flags = MMC_RSP_SPI_R2 | MMC_RSP_R1 | MMC_CMD_AC;
	cmd.flags = MMC_RSP_R1 | MMC_CMD_AC;	//zhf: mark SPI mode temporarily by James Tian

	err = mmc1_wait_for_cmd(card->host, &cmd, MMC_CMD_RETRIES);
	if (err) {
		DBG("[%s] e1\n",__func__);
		return err;
	}
	/* NOTE: callers are required to understand the difference
	 * between "native" and SPI format status words!
	 */
	if (status)
		*status = cmd.resp[0];

	DBG("[%s] e2\n",__func__);
	return 0;
}

